local addonName, AT = ...

-- file import
local Settings = AT.Settings
local Utils = AT.Utils

-- emum import
local BUFF_INFO = AT.Enums.BUFF_INFO
local SCAN_CATEGORY = AT.Enums.SCAN_CATEGORY
local TRANSPARENCY_STATE = AT.Enums.TRANSPARENCY_STATE

-- consts
local ICON_SIZE = 13
local EXPIRATION_TIME_EXTRA_CHARS = (' - '):len()

-- "class" setup
local FrameOverview = {}
setmetatable(FrameOverview, {__index = AT.FrameBase})
FrameOverview.frame = nil
FrameOverview.padding = 8
FrameOverview.dragable = true
FrameOverview.allowedTransparencyStates = {
  TRANSPARENCY_STATE.NO_MANA
}

FrameOverview.strTable = {} -- use always one entry per line
FrameOverview.approxAvgLetterWidth = nil
FrameOverview.tripleDotWidth = nil
FrameOverview.iconWidth = nil

AT.FrameOverview = FrameOverview


function FrameOverview:construct(aId)
  obj = {}
  setmetatable(obj, {__index = self})
  obj:postConstruct(aId)
  obj:Init()
  return obj
end

function FrameOverview:Init()
  self.frame = CreateFrame('FRAME', nil, UIParent)

  -- enable mouse input for spell tooltips
  self.frame:EnableMouse(true)
  self.frame:SetHyperlinksEnabled(true)
  self.frame:SetScript('OnHyperlinkClick', ChatFrame_OnHyperlinkShow)

  -- text
  self.frame.text = self.frame:CreateFontString(nil, 'BACKGROUND', GameFontNormal)
  self.frame.text:SetPoint('TOPLEFT', self.padding, -self.padding)
  self.frame.text:SetJustifyV('TOP')
  self.frame.text:SetJustifyH('LEFT')
  self.frame.text:SetFont('Fonts\\FRIZQT__.TTF', 10, 'OUTLINE')
  self.frame.text:SetNonSpaceWrap(false)

  -- calculate sizes required for max width
  local str = ' abcdefghijklmnonpqrstuvwxyz'
  self.frame.text:SetText(str)
  self.approxAvgLetterWidth = self.frame.text:GetStringWidth() / str:len()
  self.frame.text:SetText('...')
  self.tripleDotWidth = self.frame.text:GetStringWidth()
  local _, _, icon = GetSpellInfo(AT.Constants.SPELL_ID_SPELLSTEAL)
  self.frame.text:SetText('|T'..icon..':'..ICON_SIZE..'|t')
  self.iconWidth = self.frame.text:GetStringWidth()
  self.frame.text:SetText('')

  self:PostInit()
  self:Hide()
end

-- TODO: move to base class
function FrameOverview:Hide()
  wipe(self.strTable)
  self.isVisible = false
  self.frame:Hide()
end

function FrameOverview:Add(aTargetUnit, aStealableBuffsInfo, aSearchCategory)
  if not self:GetEnabled() then
    return
  end

  local emptyOrLineBreak = ''
  local maxWidth = Settings.frameOverviewMaxWidth

  local function AdjustStrWidth(aStr)
    if aStr:len() * self.approxAvgLetterWidth > maxWidth then
      return aStr:sub(0, math.floor((maxWidth - self.tripleDotWidth) / self.approxAvgLetterWidth))..'...'
    end
    return aStr
  end

  -- add line breaks if not first entry
  if #self.strTable > 0 then
    self.strTable[#self.strTable + 1] = '\n'
    emptyOrLineBreak = '\n'
  end

  -- add how unit was found
  if Settings.frameOverviewDisplayUnitOrigin then
    self.strTable[#self.strTable + 1] = string.format('%s|c%s%s|r', emptyOrLineBreak, self.textColors.unitOrigin, AdjustStrWidth(SCAN_CATEGORY:ToStr(aSearchCategory)))
    emptyOrLineBreak = '\n'
  end

  -- add unit name
  self.strTable[#self.strTable + 1] = string.format('%s|c%s%s|r', emptyOrLineBreak, self.textColors.unitName, AdjustStrWidth(aTargetUnit))
  emptyOrLineBreak = ''

  -- add stealable buffs
  for _, buffInfo in ipairs(aStealableBuffsInfo) do
    if buffInfo[BUFF_INFO.SHOW_IN_OVERVIEW] then
      -- calc expiration time lenght
      local timeSuffix = ''
      local timeSuffixLen = 0
      if buffInfo[BUFF_INFO.EXPIRATION_TIME] ~= nil and buffInfo[BUFF_INFO.EXPIRATION_TIME] > 0 then
        local formattedTime = Utils:FormatExpirationTime(buffInfo[BUFF_INFO.EXPIRATION_TIME])
        timeSuffix = '|c'..self.textColors.buffExpTime..' - '..formattedTime..'|r'
        timeSuffixLen = formattedTime:len() + EXPIRATION_TIME_EXTRA_CHARS
      end
      -- calc spellname lenght
      local spellName = buffInfo[BUFF_INFO.NAME]
      local spellNameLen = spellName:len()
      if (spellNameLen + timeSuffixLen) * self.approxAvgLetterWidth > maxWidth then
        spellName = spellName:sub(0, math.floor((maxWidth - self.tripleDotWidth) / self.approxAvgLetterWidth) - timeSuffixLen)..'...'
      end
      -- merge string parts
      self.strTable[#self.strTable + 1] = string.format('\n|Hspell:%s|h|T%s:%s|t|c%s%s%s|h',
        buffInfo[BUFF_INFO.ID], buffInfo[BUFF_INFO.ICON], ICON_SIZE, self.textColors.buffName, spellName, timeSuffix)
    end
  end
end

function FrameOverview:Finalize()
  local maxHeight = Settings.frameOverviewMaxHeight

  -- concat string and set text
  self.frame.text:SetText(table.concat(self.strTable))

  -- set window size
  self.frame.text:SetSize(9999, 9999)
  local width = self.frame.text:GetStringWidth() + self.padding * 1.5
  local height = self.frame.text:GetStringHeight()
  -- cut down string if it is too long
  if height > maxHeight then
    local maxIndex = math.floor(#self.strTable * (maxHeight / height))
    for i = #self.strTable, maxIndex, -1 do
      tremove(self.strTable, i)
    end

    local tripleDotIndex = self.strTable[#self.strTable] == '\n'
      and #self.strTable
      or #self.strTable + 1
    self.strTable[tripleDotIndex] = '\n( . . . )'

    self.frame.text:SetText(table.concat(self.strTable))
    height = self.frame.text:GetStringHeight()
  end

  -- apply min width
  width = math.max(width, Settings.frameOverviewMinWidth)
  height = math.max(height, Settings.frameOverviewMinHeight)

  self.frame:SetSize(width, height + (self.padding * 2) + 1)
  -- add some extra space to fontstring, just to make sure that is does not overflow and causes linebreaks
  self.frame.text:SetSize(width * 1.2, height * 1.2)

  wipe(self.strTable)
  self.isVisible = true
  self.frame:Show()
end
