local addonName, AT = ...
local DEBUG = false
local function _assert(aCnd, aMsg) if DEBUG then assert(aCnd, aMsg) end return aCnd end

----------------------------------------------------------------------
--           Basic declaration and initialisiation
----------------------------------------------------------------------

----------------------------
--        Imports
----------------------------

local Constants = AT.Constants
local nameplateMap = AT.nameplateMap
local NameplateFrames = AT.NameplateFrames
local Settings = AT.Settings
local Strings = AT.Strings
local TestData = AT.TestData
local Utils = AT.Utils

----------------------------
--         Enums
----------------------------

local BUFF_INFO = AT.Enums.BUFF_INFO
local BUFF_STOLEN_CHAT_MESSAGE = AT.Enums.BUFF_STOLEN_CHAT_MESSAGE
local PROFILE_TYPE = AT.Enums.PROFILE_TYPE
local SCAN_CATEGORY = AT.Enums.SCAN_CATEGORY
local SCAN_RESULT = AT.Enums.SCAN_RESULT
local SCAN_RETURN_VALUE = AT.Enums.SCAN_RETURN_VALUE
local TRANSPARENCY_STATE = AT.Enums.TRANSPARENCY_STATE

----------------------------
--      Constants
----------------------------

local CONTINUOUSLY_RESCAN_INTERVAL = 1
local RANGE_CHECK_INTERVAL = 1
local SPELL_ID_SPELLSTEAL = Constants.SPELL_ID_SPELLSTEAL
local GROUP_BUFFS_IDS = {
  1459, -- Arcane Intellect
  6673, -- Battle Shout
  133539, -- Mark of the Wild
  21562, -- Power Word Fortitude
}
local BLACKLISTED_BUFFS = {
  642, -- Divine Shield
  45438, -- Ice Block
}

----------------------------
--        States
----------------------------

local Player_GUID = nil
local Is_In_Combat = false
local Current_Target = nil
local Current_Focus = nil
local Last_Target = nil
local Last_Target_Change = 0
local Nameplate_Missing = true
local Spell_Name_Spellsteal = nil
local Group_Buff_Names = {}
local No_Mana_Spellsteal = nil

----------------------------
--      Cached CVars
----------------------------

local CVar_UNIT_NAMEPLATES_AUTOMODE = nil

----------------------------
--     Main frame init
----------------------------

local Main = CreateFrame('frame')
Main.useTestData = false
Main.npTestSettings = {}
Main.playerPosOnEnablingTestData = {}
Main.timeSinceLastUpdate = 0
Main.timeSinceLastRangeCheck = 0
Main.lastSoundPlayedTimes = {}
Main.scannedUnitGuids = {}
Main.unitGuidUnitIdMap = {}
Main.lastScanResults = {}
Main.lastScanResultsContainsNp = false
Main.unitsWithStealableBuff = {}
Main.stealableBuffs = {}
Main.frameTarget = nil
Main.frameFocus = nil
Main.frameMouseover = nil
Main.frameOverview = nil
Main.userFilterGlobalIds = {}
Main.userFilterGlobalNames = {}
Main.userfilterGlobalCsvIsBlacklist = nil
Main.userFilterContextualIds = {}
Main.userFilterContextualNames = {}
Main.userFilterContextualCsvIsBlacklist = nil
Main.userFilterContextualLua = nil
Main.userFilterContextualPrintedErrors = {}
Main:SetScript('OnEvent', function(self, event, ...) self[event](self, ...) end)
Main:RegisterEvent('PLAYER_LOGIN')
AT.Main = Main

----------------------------------------------------------------------
--                         Game Events
----------------------------------------------------------------------

function Main:PLAYER_LOGIN()
  self:UnregisterEvent('PLAYER_LOGIN')
  self['PLAYER_LOGIN'] = nil

  self:Setup()
end

function Main:CVAR_UPDATE(aCVar, aValue)
  if aCVar == 'UNIT_NAMEPLATES_SHOW_ENEMIES' and tonumber(aValue) == 0 then
    -- show nameplates disabled warning
    if Settings.npEnabled or (Settings.frameOverviewEnabled and Settings.filterOverviewShowNameplates) then
      print(Constants.ADDON_DISPLAY_NAME..": |cffffff00It is recommended to turn nameplates for enemy units on in order to use all enabled addon features.|r")
    end
  elseif aCVar == 'UNIT_NAMEPLATES_AUTOMODE' then
    CVar_UNIT_NAMEPLATES_AUTOMODE = tonumber(aValue)
  end
end

function Main:PLAYER_TARGET_CHANGED()
  Last_Target = Current_Target
  Current_Target = UnitGUID('target')
  Last_Target_Change = GetTime()
  if self.frameTarget.isVisible or self:UnitHasStealableBuff('target') then
    self:Rescan()
  end
end

function Main:PLAYER_FOCUS_CHANGED()
  Current_Focus = UnitGUID('focus')
  if self.frameFocus.isVisible or self:UnitHasStealableBuff('target') then
    self:Rescan()
  end
end

function Main:UPDATE_MOUSEOVER_UNIT()
  self:RescanConditional('mouseover')
end

function Main:NAME_PLATE_UNIT_ADDED(aUnit)
  -- add to nameplate map
  nameplateMap[aUnit] = C_NamePlate.GetNamePlateForUnit(aUnit)

  self:RescanConditional(aUnit)
end

function Main:NAME_PLATE_UNIT_REMOVED(aUnit)
  -- clean up nameplate
  local np = nameplateMap[aUnit]
  -- can be nil when loading the addon later as a non mage while a unit is visible and either
  -- "Always show nameplates" is enabled or the unit is is involved in a combat with the player
  if np then
    NameplateFrames:Clear(np)
  end

  -- remove from nameplate map
  nameplateMap[aUnit] = nil

  self:RescanConditional(aUnit, true)
end

function Main:SPELL_UPDATE_USABLE()
  -- update no mana transparency
  local spellStealUsable, noMana = IsUsableSpell(SPELL_ID_SPELLSTEAL)
  if noMana ~= No_Mana_Spellsteal then
    No_Mana_Spellsteal = noMana

    local mode = TRANSPARENCY_STATE.NO_MANA
    self.frameTarget:ToggleTransparency(mode, noMana)
    self.frameFocus:ToggleTransparency(mode, noMana)
    self.frameOverview:ToggleTransparency(mode, noMana)
    self.frameMouseover:ToggleTransparency(mode, noMana)
  end
end

function Main:PLAYER_REGEN_DISABLED()
  Is_In_Combat = true
end

function Main:PLAYER_REGEN_ENABLED()
  Is_In_Combat = false
end

function Main:MODIFIER_STATE_CHANGED(aKey, aNewState)
  -- enable / disable UI dragging on alt key
  if Settings.frameDragWithModifierKey and aKey == 'LALT' then
    self.frameTarget:ToggleDragging(aNewState == 1)
    self.frameFocus:ToggleDragging(aNewState == 1)
    self.frameOverview:ToggleDragging(aNewState == 1)
  end
end

function Main:COMBAT_LOG_EVENT_UNFILTERED()
  local timestamp, event, hideCaster, sourceGUID, sourceName, sourceFlags, sourceRaidFlags, destGUID, destName, destFlags, destRaidFlags = CombatLogGetCurrentEventInfo()

  -- print spell stolen
  if event == 'SPELL_STOLEN' then
    if sourceGUID == Player_GUID then
      local spellId, spellName, spellSchool, amount = select(12, CombatLogGetCurrentEventInfo())
        if Settings.buffStolenChatMessage == BUFF_STOLEN_CHAT_MESSAGE.PRIVATE then
          print("Spell stolen: "..GetSpellLink(amount).." from "..destName)
        elseif Settings.buffStolenChatMessage == BUFF_STOLEN_CHAT_MESSAGE.PUBLIC then
          local index = GetChannelName('General')
          if index ~= nil then 
            SendChatMessage("stole "..GetSpellLink(amount).." from "..destName, 'SAY', nil, index)
          end
        end
        if DEBUG then
          print(amount)
        end
    end
  end

  -- may rescan on aura added
  if event == 'SPELL_AURA_APPLIED' then
    if bit.band(destFlags, COMBATLOG_OBJECT_REACTION_FRIENDLY) == 0 then
      local unitId = self.unitGuidUnitIdMap[destGUID]
      if unitId then
        self:RescanConditional(unitId)
      end
    end
  end

  -- rescan on stealable aura removed
  if event == 'SPELL_AURA_REMOVED' then
    if bit.band(sourceFlags, COMBATLOG_OBJECT_REACTION_FRIENDLY) == 0 then
      local spellId, spellName, spellSchool, amount = select(12, CombatLogGetCurrentEventInfo())
      if tContains(self.stealableBuffs, spellId) and tContains(self.unitsWithStealableBuff, sourceGUID) then
        self:Rescan()
      end
    end
  end
end

----------------------------------------------------------------------
--                        Main Logic
----------------------------------------------------------------------

function Main:Setup(aForced)
  -- disable addon for non mages
  local _, classFile = UnitClass('player')
  if classFile ~= 'MAGE' and not aForced then
    AT.FrameSettings:Init(false)
    return
  end

  --init general stuff
  Player_GUID = UnitGUID('player')
  Spell_Name_Spellsteal = GetSpellInfo(Constants.SPELL_ID_SPELLSTEAL)
  for i, v in ipairs(GROUP_BUFFS_IDS) do
    local name = GetSpellInfo(v)
    tinsert(Group_Buff_Names, name)
  end

  -- init settings
  Settings:Init()
  -- part of the workaround to use a setting to update the "show test frames" checkbox.
  -- wrapped in if block to don't generated unecessary data that must get serialized
  if Settings.testFramesEnabled then
    Settings.testFramesEnabled = false
  end

  -- create frames
  self.frameTarget = AT.FrameImportant:construct('frameTarget', "TARGET")
  self.frameFocus = AT.FrameImportant:construct('frameFocus', "FOCUS")
  self.frameMouseover = AT.FrameMouseover:construct('frameMouseover')
  self.frameOverview = AT.FrameOverview:construct('frameOverview')

  -- init option frame
  AT.FrameSettings:Init(true)
  -- init nameplate frame handler
  NameplateFrames:Init()

  -- add listener for profile change
  Settings.OnProfileChanged:Add(self, self.OnProfileChanged, PROFILE_TYPE.GENERAL + PROFILE_TYPE.FILTER)

  --update custom filter
  self:UpdateCustomFilters()

  -- register game events
  self:RegisterEvent('CVAR_UPDATE')
  self:RegisterEvent('PLAYER_FOCUS_CHANGED')
  self:RegisterEvent('PLAYER_TARGET_CHANGED')
  self:RegisterEvent('SPELL_UPDATE_USABLE')
  self:RegisterEvent('NAME_PLATE_UNIT_ADDED')
  self:RegisterEvent('NAME_PLATE_UNIT_REMOVED')
  self:RegisterEvent('PLAYER_REGEN_ENABLED')
  self:RegisterEvent('PLAYER_REGEN_DISABLED')
  self:RegisterEvent('UPDATE_MOUSEOVER_UNIT')
  self:RegisterEvent('MODIFIER_STATE_CHANGED')
  self:RegisterEvent('COMBAT_LOG_EVENT_UNFILTERED')
  self:SetScript('OnUpdate', self.OnUpdate)
  InterfaceOptionsFrame:HookScript('OnHide', function() Main:OnInterfaceOptionsClosed() end)

  -- trigger cvar updates
  self:CVAR_UPDATE('UNIT_NAMEPLATES_SHOW_ENEMIES', GetCVar('nameplateShowEnemies'))
  self:CVAR_UPDATE('UNIT_NAMEPLATES_AUTOMODE', GetCVar('nameplateShowAll'))

  -- show welcome message
  if Settings.showWelcomeMsg then
    Settings.showWelcomeMsg = false
    Utils:Print(Strings.welcomeMsg)
  end
end

function Main:OnUpdate(aElapsed)
  -- update frame mouseover position
  if self.frameMouseover.isVisible then
    if UnitExists('mouseover') or (self.useTestData and Settings.testFramesShowMouseover) then
      self.frameMouseover:OnUpdate()
    else
      self.frameMouseover:Hide()
    end
  end

  -- range check every x seconds
  self.timeSinceLastRangeCheck = self.timeSinceLastRangeCheck + aElapsed
  if self.timeSinceLastRangeCheck > RANGE_CHECK_INTERVAL then
    self.timeSinceLastRangeCheck = 0
    -- target
    if self.frameTarget.isVisible then
      local isInRange = IsSpellInRange(Spell_Name_Spellsteal, 'target')
      if isInRange ~= nil then
        self.frameTarget:ToggleTransparency(TRANSPARENCY_STATE.NOT_IN_RANGE, tonumber(isInRange) == 0)
      end
    end
    -- focus
    if self.frameFocus.isVisible then
      local isInRange = IsSpellInRange(Spell_Name_Spellsteal, 'focus')
      if isInRange ~= nil then
        self.frameFocus:ToggleTransparency(TRANSPARENCY_STATE.NOT_IN_RANGE, tonumber(isInRange) == 0)
      end
    end
  end

  -- continuously rescan every x seconds
  self.timeSinceLastUpdate = self.timeSinceLastUpdate + aElapsed
  if self.timeSinceLastUpdate > CONTINUOUSLY_RESCAN_INTERVAL then
    self.timeSinceLastUpdate = 0
    self:Rescan()
  end

  -- disable test frames when player starts walking
  if self.useTestData then
    local posX, posY = UnitPosition('player')
    if ((self.playerPosOnEnablingTestData.x - posX) ^ 2 + (self.playerPosOnEnablingTestData.y - posY) ^ 2) ^ 0.5 > 2.5 then
      self:SetTestFramesEnabled(false)
    end
  end
end

function Main:OnProfileChanged(aProfileType)
  if aProfileType == PROFILE_TYPE.FILTER then
    self:UpdateCustomFilters()
  end

  self:RescanNextTick()
end

function Main:OnInterfaceOptionsClosed()
  if not self.useTestData or not Settings.testFramesShowNameplateIcons then
    self:RestoreForceShowNameplates(self.npTestSettings, true)
  end
end

-- @RETURNS Boolean
function Main:RescanConditional(aUnit, aIsRemovedUnit)
  if not Is_In_Combat and not Settings.scanOutOfCombat then
    return false
  end
  if (aIsRemovedUnit and tContains(self.unitsWithStealableBuff, aUnit)) 
    or (not aIsRemovedUnit and self:UnitHasStealableBuff(aUnit)) then

    self:Rescan()
    return true
  end
  return false
end

function Main:RescanNextTick()
  self.timeSinceLastUpdate = CONTINUOUSLY_RESCAN_INTERVAL
end

-- scan for buffs that can be stolen
function Main:Rescan()
  -- clear cached stuff
  wipe(self.scannedUnitGuids)
  wipe(self.unitGuidUnitIdMap)
  wipe(self.unitsWithStealableBuff)
  wipe(self.stealableBuffs)

  -- reset timer for continuously rescan
  self.timeSinceLastUpdate = 0
  -- set range check timer to max
  self.timeSinceLastRangeCheck = RANGE_CHECK_INTERVAL

  -- clear all nameplates
  NameplateFrames:ClearAll()

  -- const
  local frameTargetEnabled = self.frameTarget:GetEnabled()
  local frameFocusEnabled = self.frameFocus:GetEnabled()
  local frameOverviewEnabled = self.frameOverview:GetEnabled()
  local npEnabled = Settings.npEnabled

  local scanResults = {}
  if self.useTestData then
    -- use testdata
    scanResults = TestData.data

    -- fill nameplates
    if Settings.testFramesShowNameplateIcons then
      TestData:FillNameplates()
    end
  elseif Is_In_Combat or Settings.scanOutOfCombat then
    --[[
      Scan Units. 
      Units that are only in in a spefic context releant (mostly overiew)
      will be filtered out here for performce reasons 
    ]]

    -- scan target
    self:TryScanUnit('target', SCAN_CATEGORY.TARGET, scanResults)

    -- scan focus
    self:TryScanUnit('focus', SCAN_CATEGORY.FOCUS, scanResults)

    -- scan pet target
    if frameOverviewEnabled and Settings.filterOverviewShowPetTarget then
      self:TryScanUnit('pet-target', SCAN_CATEGORY.PET_TARGET, scanResults)
    end

    -- scan bosses
    if frameOverviewEnabled and Settings.filterOverviewShowBosses then
      for bossIndex = 1, 9 do
        if self:TryScanUnit('boss'..bossIndex, SCAN_CATEGORY.BOSS, scanResults) == SCAN_RETURN_VALUE.INVALID_UNIT then
          break
        end
      end
    end

    -- scan nameplates
    if npEnabled or (frameOverviewEnabled and Settings.filterOverviewShowNameplates) then
      for k, _ in pairs(nameplateMap) do
        self:TryScanUnit(k, SCAN_CATEGORY.NAMEPLATE, scanResults)
      end
    end

    local scanGroupMemberTargets = frameOverviewEnabled and Settings.filterOverviewShowGroupMemberTargets
    local scanGroupMemberPetTargets = frameOverviewEnabled and Settings.filterOverviewShowGroupMemberPetTargets
    if scanGroupMemberTargets or scanGroupMemberPetTargets then
      local groupType = nil
      if IsInGroup() then
        groupType = IsInRaid() and 'raid' or 'party'
      end
      if groupType then
        for partyMember = 1, GetNumGroupMembers() do
          -- scan group member targets
          if scanGroupMemberTargets then
            local unit = groupType..partyMember
            local inRange, _ = UnitInRange(unit)
            if inRange then
              self:TryScanUnit(unit..'-target', SCAN_CATEGORY.GROUP_MEMBER_TARGET, scanResults)
            end
          end

          -- scan group member pet targets
          if scanGroupMemberPetTargets then
            local unit = groupType..partyMember..'-pet'
            local inRange, _ = UnitInRange(unit)
            if inRange then
              self:TryScanUnit(unit..'-target', SCAN_CATEGORY.GROUP_MEMBER_PET_TARGET, scanResults)
            end
          end
        end
      end
    end

    -- scan mouseover
    if Settings.frameMouseoverEnabled then
      self:TryScanUnit('mouseover', SCAN_CATEGORY.MOUSEOVER, scanResults, true)
    end
  end


  -- show in UI

  -- const
  local targetCount = #scanResults
  local firstResult = scanResults[1]
  local lastFrameTargetVisible = self.frameTarget.isVisible
  local lastFrameFocusVisible = self.frameFocus.isVisible
  local lastFrameMouseoverVisible = self.frameMouseover.isVisible
  local lastFrameOverviewVisible = self.frameOverview.isVisible
  -- variable
  local targetHasStealableBuff = false
  local focusHasStealableBuff = false
  local overviewHasStealableBuff = false
  local npHasStealableBuff = false
  local mouseoverHasStealableBuff = false
  local nonImportantHasStealableBuff = false
  local mouseoverGuid = nil

  if targetCount > 0 then
    for _, scanResult in ipairs(scanResults) do
      local category = scanResult[SCAN_RESULT.CATEGORY]

      -- target
      if frameTargetEnabled and category == SCAN_CATEGORY.TARGET and scanResult[SCAN_RESULT.HAS_IMPORTANT_BUFFS] then
        self.frameTarget:Set(scanResult[SCAN_RESULT.UNIT], scanResult[SCAN_RESULT.BUFFS_INFO])
        targetHasStealableBuff = true
        -- gets set here too, because otherwise np sound would be played 
        -- in next rescan when nameplate gets big through targeting
        self.lastScanResultsContainsNp = true
      -- focus
      elseif frameFocusEnabled and category == SCAN_CATEGORY.FOCUS and scanResult[SCAN_RESULT.HAS_IMPORTANT_BUFFS] then
        self.frameFocus:Set(scanResult[SCAN_RESULT.UNIT], scanResult[SCAN_RESULT.BUFFS_INFO])
        focusHasStealableBuff = true
      --nothing above
      else
        nonImportantHasStealableBuff = true
      end

      -- mouseover
      if category == SCAN_CATEGORY.MOUSEOVER and scanResult[SCAN_RESULT.HAS_MOUSEOVER_BUFFS] then
        mouseoverGuid = scanResult[SCAN_RESULT.UNIT_GUID]
        if mouseoverGuid ~= Current_Target or self.useTestData then
          self.frameMouseover:Set(scanResult[SCAN_RESULT.BUFFS_INFO])
          mouseoverHasStealableBuff = true
        end
      end

      -- overview
      if frameOverviewEnabled and scanResult[SCAN_RESULT.HAS_OVERVIEW_BUFFS] then
        self.frameOverview:Add(self.useTestData and scanResult[SCAN_RESULT.UNIT] or GetUnitName(scanResult[SCAN_RESULT.UNIT]),
          scanResult[SCAN_RESULT.BUFFS_INFO], category)
        overviewHasStealableBuff = true
      end

      -- nameplates
      if npEnabled and scanResult[SCAN_RESULT.HAS_NAMEPLATE_BUFFS] then
        local np = C_NamePlate.GetNamePlateForUnit(scanResult[SCAN_RESULT.UNIT])
        if np then
          NameplateFrames:Add(np, scanResult[SCAN_RESULT.BUFFS_INFO])
          npHasStealableBuff = true
        end
      end
    end
  end

  -- order for the following it relevant for the sound prioritization
  self:SoundSpellFoundResetCache()

  -- target
  if targetHasStealableBuff then
    if not lastFrameTargetVisible then
      self:SoundSpellFoundPlay('Target')
    end
  else
    self.frameTarget:Hide()
  end

  -- focus
  if focusHasStealableBuff then
    if not lastFrameFocusVisible then
      self:SoundSpellFoundPlay('Focus')
    end
  else
    self.frameFocus:Hide()
  end

  -- mouseover
  if mouseoverHasStealableBuff then
    if not lastFrameMouseoverVisible then
      self:SoundSpellFoundPlay('Mouseover')
    end
  else
    self.frameMouseover:Hide()
  end

  -- overview
  if overviewHasStealableBuff and nonImportantHasStealableBuff
    -- workaround for the "issue" that target gets a nameplate for a few moments when untargeting via tabbing (only relevant when not in combat, causes short overview popup)
    and (CVar_UNIT_NAMEPLATES_AUTOMODE == 1 or not (targetCount == 1 and Is_In_Combat == false and not UnitAffectingCombat(firstResult[SCAN_RESULT.UNIT]) 
    and GetTime() < Last_Target_Change + 1 and firstResult[SCAN_RESULT.UNIT_GUID] == Last_Target and Utils:StrStartsWith(firstResult[SCAN_RESULT.UNIT], 'nameplate'))) then

    self.frameOverview:Finalize()

    if not lastFrameOverviewVisible then
      -- workaround for sound after switching away from target with stealable buffs
      if not (targetCount == 1 and lastFrameTargetVisible and not self.frameTarget.isVisible) then
        self:SoundSpellFoundPlay('Overview')
      end
    end
  else
    self.frameOverview:Hide()
  end

  -- nameplates
  if npHasStealableBuff and not self.lastScanResultsContainsNp then
    self:SoundSpellFoundPlay('Np')
  end

  self.lastScanResults = scanResults
  self.lastScanResultsContainsNp = npHasStealableBuff
end

-- @RETURNS SCAN_RESULT
function Main:TryScanUnit(aUnit, aCategory, --[[out]] outScanResult, allowDuplicate)
  if not UnitExists(aUnit) then
    return SCAN_RETURN_VALUE.INVALID_UNIT
  end

  local returnValue = SCAN_RETURN_VALUE.NO_SPELLS_FOUND
  local guid = UnitGUID(aUnit)

  -- add unit to guid -> id map
  if self.unitGuidUnitIdMap[guid] == nil then
    self.unitGuidUnitIdMap[guid] = aUnit
  end

  if (allowDuplicate or not tContains(self.scannedUnitGuids, guid)) and self:FilterUnit(aUnit) then
    -- iterate over all buffs and write out stealable buffs
    local aStealableBuffsInfo = {}
    local addedBuffs = {}
    local hasImportantEntries = false
    local hasMouseoverBuffs = false
    local hasOverviewEntries = false
    local hasNameplateEntries = false

    for buffIndex = 1, 999 do
      local name, icon, count, debuffType, duration, expirationTimeRaw, unitCaster, canStealOrPurge, nameplateShowPersonal, spellId, canApplyAura,
        isBossDebuff, isCastByPlayer, nameplateShowAll = UnitAura(aUnit, buffIndex, 'HELPFUL')

      if not name then
        break
      end

      if expirationTimeRaw ~= nil and self:FilterGlobal(spellId, name, canStealOrPurge) then
        local expirationTime = expirationTimeRaw > 0 and math.floor(expirationTimeRaw - GetTime() + 0.5) or nil

        -- prepare lua filter
        local args = {}
        if self:HasFilterContextualLua() then
          args.unitInfo = {}
          args.unitInfo.id = aUnit
          args.unitInfo.guid = guid
          args.buffInfo = {}
          args.buffInfo.name = name
          args.buffInfo.count = count
          args.buffInfo.buffType = debuffType
          args.buffInfo.expirationTime = expirationTime
          args.buffInfo.expirationTimeRaw = expirationTimeRaw
          args.buffInfo.unitCaster = unitCaster
          args.buffInfo.spellId = spellId
          args.buffInfo.canApplyAura = canApplyAura
          args.buffInfo.isBossBuff = isBossDebuff
          args.buffInfo.isCastByPlayer = isCastByPlayer
          args.buffInfo.nameplateShowAll = nameplateShowAll
        end

        -- call built-in and user filters
        local isImportantEntry, isMouseoverEntry, isOverviewEntry, isNameplateEntry = self:FilterContextual(aCategory, spellId, name, expirationTime, args)

        if not tContains(addedBuffs, spellId) then
          local stealableBuffInfo = {}
          stealableBuffInfo[BUFF_INFO.ID] = spellId
          stealableBuffInfo[BUFF_INFO.NAME] = name
          stealableBuffInfo[BUFF_INFO.ICON] = icon
          stealableBuffInfo[BUFF_INFO.COUNT] = count
          stealableBuffInfo[BUFF_INFO.EXPIRATION_TIME] = expirationTime
          stealableBuffInfo[BUFF_INFO.SHOW_IN_MOUSEOVER] = isMouseoverEntry
          stealableBuffInfo[BUFF_INFO.SHOW_IN_OVERVIEW] = isOverviewEntry
          stealableBuffInfo[BUFF_INFO.SHOW_IN_NAMEPLATE] = isNameplateEntry
          tinsert(aStealableBuffsInfo, stealableBuffInfo)
          tinsert(addedBuffs, spellId)
        else
          -- merge multiple times applied, but not stacked, buffs
          for _, v in ipairs(aStealableBuffsInfo) do
            if v[BUFF_INFO.ID] == spellId then
              local cnt = v[BUFF_INFO.COUNT]
              v[BUFF_INFO.COUNT] = cnt == 0 and 2 or cnt + (count == 0 and 1 or count)
              local expTime = v[BUFF_INFO.EXPIRATION_TIME]
              v[BUFF_INFO.EXPIRATION_TIME] = AT.Utils:Ternary(expTime ~= nil and expirationTime > expTime, expirationTime, expTime)
              v[BUFF_INFO.SHOW_IN_MOUSEOVER] = AT.Utils:Ternary(v[BUFF_INFO.SHOW_IN_MOUSEOVER], true, isMouseoverEntry)
              v[BUFF_INFO.SHOW_IN_OVERVIEW] = AT.Utils:Ternary(v[BUFF_INFO.SHOW_IN_OVERVIEW], true, isOverviewEntry)
              v[BUFF_INFO.SHOW_IN_NAMEPLATE] = AT.Utils:Ternary(v[BUFF_INFO.SHOW_IN_NAMEPLATE], true, isNameplateEntry)
              break
            end
          end
          if DEBUG then
            print('merged', spellId)
          end
        end

        -- note if unit has buffs which we want to show in specific frames
        if isImportantEntry then
          hasImportantEntries = true
        end
        if isMouseoverEntry then
          hasMouseoverBuffs = true
        end
        if isOverviewEntry then
          hasOverviewEntries = true
        end
        if isNameplateEntry then
          hasNameplateEntries = true
        end

        -- add to stealable buffs list
        if not tContains(self.stealableBuffs, spellId) then
          tinsert(self.stealableBuffs, spellId)
        end
      end
    end

    -- store result in return value table
    if #aStealableBuffsInfo > 0 then
      local scanResult = {}
      scanResult[SCAN_RESULT.UNIT] = aUnit
      scanResult[SCAN_RESULT.UNIT_GUID] = guid
      scanResult[SCAN_RESULT.CATEGORY] = aCategory
      scanResult[SCAN_RESULT.BUFFS_INFO] = aStealableBuffsInfo
      scanResult[SCAN_RESULT.HAS_IMPORTANT_BUFFS] = hasImportantEntries
      scanResult[SCAN_RESULT.HAS_MOUSEOVER_BUFFS] = hasMouseoverBuffs
      scanResult[SCAN_RESULT.HAS_OVERVIEW_BUFFS] = hasOverviewEntries
      scanResult[SCAN_RESULT.HAS_NAMEPLATE_BUFFS] = hasNameplateEntries

      tinsert(self.unitsWithStealableBuff, guid)
      tinsert(self.unitsWithStealableBuff, aUnit)

      tinsert(outScanResult, scanResult)
      returnValue = SCAN_RETURN_VALUE.SPELLS_FOUND
    end
  end

  -- note that unit was scanned
  self.scannedUnitGuids[#self.scannedUnitGuids + 1] = guid

  return returnValue
end

-- Determines wether a unit should get scanned or not
-- @RETURNS Boolean
function Main:FilterUnit(aUnit)
  return UnitCanAttack('player', aUnit) and (not Settings.filterIgnoreUnitsNotAffectingCombat or UnitAffectingCombat(aUnit))
end

-- Determines wether a buff should be shown in general. If false, the spell
-- will be shown nowhere, even not if FilterContextual would return true
-- @RETURNS Boolean
function Main:FilterGlobal(aSpellId, aSpellName, aCanStealOrPurge)
  if not aCanStealOrPurge and self.userfilterGlobalCsvIsBlacklist then
    return false
  end
  if self.userfilterGlobalCsvIsBlacklist and tContains(BLACKLISTED_BUFFS, aSpellId) then
    return false
  end
  if tContains(self.userFilterGlobalIds, aSpellId) or tContains(self.userFilterGlobalNames, strlower(aSpellName)) then
    return not self.userfilterGlobalCsvIsBlacklist
  end
  return self.userfilterGlobalCsvIsBlacklist
end

-- Determines wether a buff should be shown or not in a specific context
-- NOTE: for target/focus the spell will be shown nevertheless if there is
-- at least one another spell that must be shown
-- @RETURNS Boolean isImportantEntry
-- @RETURNS Boolean isMouseoverEntry
-- @RETURNS Boolean isOverviewEntry
-- @RETURNS Boolean isNameplateEntry
function Main:FilterContextual(aUnitCategory, aSpellId, aSpellName, aSpellExpirationTime, aLuaArgs)
  local isImportantEntry = nil
  local isMouseoverEntry = nil
  local isOverviewEntry = nil
  local isNameplateEntry = nil

  -- check if its an ignored group buff
  if Settings.filterIgnoreGroupBuffs and tContains(Group_Buff_Names, aSpellName) then
    return false, false, false, false
  end

  -- check if its black or whitelisted by the user
  local userFilterContains = tContains(self.userFilterContextualIds, aSpellId) or tContains(self.userFilterContextualNames, strlower(aSpellName))
  if (userFilterContains and self.userFilterContextualCsvIsBlacklist) or (not userFilterContains and not self.userFilterContextualCsvIsBlacklist) then
    return false, false, false, false
  end

  -- run lua filter
  if not self.useTestData and not self:FilterContextualLua(aLuaArgs) then
    return false, false, false, false
  end

  local unitIsTarget = aUnitCategory == SCAN_CATEGORY.TARGET
  local unitIsFocus =  aUnitCategory == SCAN_CATEGORY.FOCUS

  -- check if its target/focus and if we want to hide important targets in frame overview
  if isOverviewEntry ~= false and ((unitIsTarget and not Settings.filterOverviewShowTarget) or (unitIsFocus and not Settings.filterOverviewShowFocus)) then
    isOverviewEntry = false
  end

  -- check if we want to hide nameplates in overview
  if isOverviewEntry ~= false and aUnitCategory == SCAN_CATEGORY.NAMEPLATE and not Settings.filterOverviewShowNameplates then
    isOverviewEntry = false
  end

  -- check expirationTime
  if aSpellExpirationTime == nil then
    isImportantEntry = true
    if isOverviewEntry == nil then
      isOverviewEntry = true
    end
    isNameplateEntry = true
  else
    isImportantEntry = (unitIsTarget or unitIsFocus) and aSpellExpirationTime >= Settings.filterImportantMinExpirationTime
    local show = aSpellExpirationTime >= Settings.filterOverviewMinExpirationTime
    if isOverviewEntry == nil then
      isOverviewEntry = show
    end
    isNameplateEntry = show
  end

  -- postprocess for mouseover
  isMouseoverEntry =  isImportantEntry or isOverviewEntry or isNameplateEntry
  if aUnitCategory == SCAN_CATEGORY.MOUSEOVER then
    isOverviewEntry = false
  end

  return isImportantEntry, isMouseoverEntry, isOverviewEntry, isNameplateEntry
end

-- @RETURNS Boolean
function Main:FilterContextualLua(aArgs)
  if not self.userFilterContextualLua then
    return true
  end
  -- execute function
  local success, res = pcall(self.userFilterContextualLua, aArgs.unitInfo, aArgs.buffInfo);
  if not success then
    message(string.format("%s custom filter Lua runtime error:\n\n%s\n", Constants.ADDON_DISPLAY_NAME, res))
    return true
  end
  -- print warning on wrong return type
  local returnType = type(res)
  if returnType ~= 'boolean' then
    local warning = string.format("|cffffff00Expected boolean as return type for Lua custom filter, but got %s.|r", returnType)
    if not tContains(self.userFilterContextualPrintedErrors, warning) then
      tinsert(self.userFilterContextualPrintedErrors, warning)
      Utils:Print(warning)
    end
  end
  -- return result, with handling for wrong type
  return res == nil or true and res
end

-- @RETURNS Boolean
function Main:HasFilterContextualLua()
  return self.userFilterContextualLua ~= nil
end

function Main:UpdateCustomFilters()
  local ids, names

  -- update csv global
  ids, names = self:ParseCustomFilter(Settings.filterGlobalCsv)
  self.userFilterGlobalIds = ids
  self.userFilterGlobalNames = names
  self.userfilterGlobalCsvIsBlacklist = Settings.filterGlobalCsvIsBlacklist
  -- print warning when using an empty whitelist filter
  if not self.userfilterGlobalCsvIsBlacklist and #self.userFilterGlobalIds == 0 and #self.userFilterGlobalNames == 0 then
    Utils:Print(string.format(Strings.warningEmptyCsvWhitelist, "hard"))
  end

  -- update csv contextual
  ids, names = self:ParseCustomFilter(Settings.filterContextualCsv)
  self.userFilterContextualIds = ids
  self.userFilterContextualNames = names
  self.userFilterContextualCsvIsBlacklist = Settings.filterContextualCsvIsBlacklist
  -- print warning when using an empty whitelist filter
  if not self.userFilterContextualCsvIsBlacklist and #self.userFilterContextualIds == 0 and #self.userFilterContextualNames == 0 then
    Utils:Print(string.format(Strings.warningEmptyCsvWhitelist, "soft"))
  end

  -- update lua contextual
  wipe(self.userFilterContextualPrintedErrors)
  self.userFilterContextualLua = nil
  local luaStr = Settings.filterContextualLua
  if not Utils:StrIsNilOrEmptyOrWhitespaceOnly(luaStr) then
    -- load lua string raw to get an error message with correct line indication
    local func, errorMsg = loadstring(luaStr)
    if not func then
      message(string.format("%s custom filter Lua error:\n\n%s\n", Constants.ADDON_DISPLAY_NAME, errorMsg))
    else
      -- load lua script again, but with injected code
      func, errorMsg = loadstring(string.format([[
        return function(unitInfo, buffInfo)
          %s
        end
      ]], luaStr), 'post-processed-code')
      if not func then
        message(errorMsg)
      else
        self.userFilterContextualLua = func()
      end
    end
  end
end

-- @RETURNS Number Table (spell ids)
-- @RETURNS String Table (spell names)
function Main:ParseCustomFilter(aStr)
  local function FinalizeToken(aStrBuilder, aIds, aNames)
    local str = strtrim(table.concat(aStrBuilder))
    if str ~= '' then
      local num = tonumber(str)
      local spellId = nil
      if num ~= nil then 
        _, _, _, _, _, _, spellId = GetSpellInfo(num)
        if spellId ~= nil then
          tinsert(aIds, num)
        end
      end
      if spellId == nil then
        tinsert(aNames, strlower(str))
      end
    end
    wipe(aStrBuilder)
  end

  -- return values
  local ids = {}
  local names = {}

  -- parser
  local strBuilder = {}
  local isInComment = false
  local ignoreNextInstr = false
  for c in gmatch(aStr, '.') do
    local lastIngoreNextInstr = ignoreNextInstr
    if c == '\\' then
      ignoreNextInstr = true
    elseif c == '(' and not ignoreNextInstr then
      isInComment = true
    elseif c == ')' and not ignoreNextInstr then
      isInComment = false
    elseif c == ',' and not isInComment and not ignoreNextInstr then
      FinalizeToken(strBuilder, ids, names)
    elseif not isInComment then
      ignoreNextInstr = false
      tinsert(strBuilder, c)
    end
    if lastIngoreNextInstr then
      ignoreNextInstr = false
    end
  end
  FinalizeToken(strBuilder, ids, names)

  return ids, names
end

-- Lightweight func for just checking if a unit has a stealable buff
-- @RETURNS Boolean
function Main:UnitHasStealableBuff(aUnit)
  if self:FilterUnit(aUnit) then
    for buffIndex = 1, 999 do
      local name, _, _, _, _, _, _, canStealOrPurge, _, spellId = UnitAura(aUnit, buffIndex, 'HELPFUL')
      if not name then
        break
      end
      if self:FilterGlobal(spellId, name, canStealOrPurge) then
        return true
      end
    end
  end
  return false
end

-- Plays a spell found sound
function Main:SoundSpellFoundPlay(aSound, --[[optional]] aForced)
  -- prevent that two sounds gets played in one rescan
  if self.spellFoundSoundPlayed and not aForced then
    return
  end

  if not Settings['sound'..aSound..'Enabled'] then
    return
  end

  local timeGap = Settings['sound'..aSound..'MinTimeGap']
  local lastPlayed = self.lastSoundPlayedTimes[aSound] or 0
  local time = GetTime()
  if aForced or time > lastPlayed + timeGap then
    PlaySoundFile(Settings['sound'..aSound..'File'], Settings['sound'..aSound..'Channel'])
    self.lastSoundPlayedTimes[aSound] = time
  end
end

function Main:SoundSpellFoundResetCache()
  self.spellFoundSoundPlayed = false
end

-- En/Disables test frames
function Main:SetTestFramesEnabled(aEnabled)
  self.useTestData = aEnabled
  Settings.testFramesEnabled = aEnabled

  if aEnabled then
    TestData:Init()
    self.playerPosOnEnablingTestData.x, self.playerPosOnEnablingTestData.y = UnitPosition('player')
    Utils:Print(Strings.testFramesEnabled)
  else
    Utils:Print(Strings.testFramesDisabled)
  end

  AT.FrameSettings:RefreshAllValues()
  self:OnNameplateTestIconsEnabledChanged()
  self:Rescan()
end

function Main:OnNameplateTestIconsEnabledChanged()
  local enabled = self.useTestData and Settings.testFramesShowNameplateIcons
  local cache = self.npTestSettings

  if enabled then
    -- force show nameplates
    self:ForceShowNameplates(cache)
    -- no nameplates found error message
    if not cache.previousEnabled then
      cache.errorTimer = C_Timer.NewTimer(2, function()
        local len = Utils:TableLen(nameplateMap)
        if len == 0 or (len == 1 and UnitGUID(Utils:TableKeys(nameplateMap)[1]) == Player_GUID) then
          Utils:Print(Strings.testFrameNoNameplatesFound)
        end
      end)
    end
  else
    -- restore force show nameplates
    self:RestoreForceShowNameplates(cache, not InterfaceOptionsFrame:IsShown())
    -- cancel no nameplate error message
    if cache.errorTimer then
      cache.errorTimer:Cancel()
    end
    cache.errorTimer = nil
  end

  cache.previousEnabled = enabled
end

function Main:ForceShowNameplates(aCache)
  if aCache.targetValues == nil then
    aCache.targetValues = {
      nameplateShowAll = 1,
      nameplateShowEnemies = 1,
      nameplateShowEnemyMinions = 1,
      nameplateShowEnemyMinus = 1,
      nameplateShowFriends = 1,
      nameplateShowFriendlyMinions = 1
    }
  end
  aCache.initalValues = aCache.initalValues or {}
  for cVar, value in pairs(aCache.targetValues) do
    -- save inital value
    if not aCache.previousEnabled then
      aCache.initalValues[cVar] = GetCVar(cVar)
    end
    -- override value
    SetCVar(cVar, value)
  end
end

function Main:RestoreForceShowNameplates(aCache, aClearInitialValues)
  if not aCache.initalValues then
    return
  end
  for cVar, value in pairs(aCache.initalValues) do
    SetCVar(cVar, value)
  end
  if aClearInitialValues then
    aCache.initalValues = nil
  end
end
